
# -*- coding:utf-8 -*-

# main starting file for use with debugger. Normally the software should be started by the treeview/__main__.py file.
import faulthandler
from treetime.treetime import TreeTime

faulthandler.enable()
TreeTime()



#
# Tis file is part of TreeTime, a tree editor and data analyser
#
# Copyright (C) GPLv3, 2015, Jacob Kanev
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

# -*- coding:utf-8 -*-

from .item import *
from textwrap import wrap
import datetime
from math import floor, ceil

class Field:
    """
    A set of instructions to view/display the content of data items.
    Fields are part of nodes, and are stored in templates.
    """

    def __init__(self, node=None, ownFields=[], childFields=[], siblingFields=[], parentFields=[], fieldType=None):
        """Initialises the class, links the source node, field type, and sets the evaluation method."""
        
        self.cache = None
        self.ownFields = ownFields
        self.siblingFields = siblingFields
        self.childFields = childFields
        self.parentFields = parentFields
        self.sourceNode = node
        self.getValue = None
        self.getString = None
        self.fieldType = fieldType
        if (self.fieldType != ""):
            self.initFieldType()

    def __deepcopy__(self, memo):
        ''' Overload of the deepcopy function, to avoid copying the node recursively.'''

        if memo is None:
            memo = {}
        
        fieldType = self.__class__
        newField = fieldType.__new__(fieldType)
        
        # cache doesn't get copied
        newField.cache = None
        
        # simple fields are copies
        newField.ownFields = copy.deepcopy(self.ownFields)
        newField.siblingFields = copy.deepcopy(self.siblingFields)
        newField.childFields = copy.deepcopy(self.childFields)
        newField.parentFields = copy.deepcopy(self.parentFields)
        newField.fieldType = copy.deepcopy(self.fieldType)
        
        # the source node is a link
        newField.sourceNode = self.sourceNode
        
        # functions are set
        if (newField.fieldType != ""):
            newField.initFieldType()
        
        memo[id(self)] = newField
        return newField

    @staticmethod
    def getFieldValue(field):
        """
        :param field: the field to get the value from
        :return: Depending on the field type, the current value. Mostly this is just the content of the "content" key,
        but with timers this is different.
        """
        if field["type"] == "timer":
            running_since = field.get("running_since")
            partial = field["content"]
            running_since = running_since and datetime.datetime.strptime(running_since,
                                                                      "%Y-%m-%d %H:%M:%S")
            if running_since:
                elapsed = datetime.datetime.now() - running_since
                return partial \
                       + elapsed.days * 24.0 \
                       + elapsed.seconds / 60.0 / 60.0 \
                       + elapsed.microseconds / 60.0 / 60.0 / 1000000.0
            else:
                return partial
        else:
            return field["content"]

    def getFieldValues(self):
        """ Gets all values of all related fields in a list.
        Order is: own fields first, then child fields, then sibling fields, then parent fields.
        Item fields of the same name have precence over tree fields.
        """

        values = []
        
        # look in own fields
        node = self.sourceNode
        if node.item is not None: # don't try to get values from the root node
            for f in self.ownFields:
                if f in node.fields:
                    values += [node.fields[f].getValue()]
                elif f in node.item.fields:
                    values += [Field.getFieldValue(node.item.fields[f])]
        
        # look in child fields
        node = self.sourceNode
        for f in self.childFields:
            for c in node.children:
                if f in c.fields:
                    values += [c.fields[f].getValue()]
                elif c.item and c.item.fields and f in c.item.fields:
                    values += [Field.getFieldValue(c.item.fields[f])]

        # look in sibling fields
        node = self.sourceNode.parent
        for f in self.siblingFields:
            for c in node.children:
                if c != self.sourceNode:
                    if f in c.fields:
                        values += [c.fields[f].getValue()]
                    elif f in c.item.fields:
                        values += [Field.getFieldValue(c.item.fields[f])]

        # look in parent fields
        node = self.sourceNode.parent
        for f in self.parentFields:
            if f in node.item.fields:
                values += [node.fields[f].getValue()]
            elif f in node.fields:
                values += [Field.getFieldValue(node.item.fields[f])]
        
        # return
        return values

    def initFieldType(self):
        
        if self.fieldType == "string":
            self.getValue = self.getValueString
            self.getString = self.getStringUnchanged
        elif self.fieldType == "sum":
            self.getValue = self.getValueSum
            self.getString = self.getStringRounded
        elif self.fieldType == "sum-time":
            self.getValue = self.getValueSum
            self.getString = self.getStringTime
        elif self.fieldType == "difference":
            self.getValue = self.getValueDifference
            self.getString = self.getStringRounded
        elif self.fieldType == "difference-time":
            self.getValue = self.getValueDifference
            self.getString = self.getStringTime
        elif self.fieldType == "mean":
            self.getValue = self.getValueMean
            self.getString = self.getStringRounded
        elif self.fieldType == "mean-percent":
            self.getValue = self.getValueMean
            self.getString = self.getStringPercent
        elif self.fieldType == "ratio":
            self.getValue = self.getValueRatio
            self.getString = self.getStringRounded
        elif self.fieldType == "ratio-percent":
            self.getValue = self.getValueRatio
            self.getString = self.getStringPercent
        elif self.fieldType == "node-name":
            self.getValue = self.getValueNodeName
            self.getString = self.getStringUnchanged
        elif self.fieldType == "node-path":
            self.getValue = self.getValueNodePath
            self.getString = self.getStringUnchanged

    def getStringPercent(self):
        if self.sourceNode and self.getValue:
            v = self.getValue()
            if v:
                whitespace = ""
                if v < 0.1:
                    whitespace = "  "
                elif v < 1.0:
                    whitespace = " "
                return whitespace + str(round(100*v)) + " %"
            else:
                return ""
        else:
            return ""

    def getStringTime(self):
        if self.sourceNode and self.getValue:
            v = self.getValue()
            if v:
                if v < 0:
                    adjust = 0.49999
                else:
                    adjust = -0.49999
                hours = round(v + adjust)
                minutes = round(60.0*(v - hours) + adjust)
                seconds = round(3600.0*((v - hours) - minutes/60.0) + adjust)
                sign = v < 0.0 and "- " or "  "
                return "{}{:02d}:{:02d}:{:02d}".format(sign, abs(hours), abs(minutes), abs(seconds))
            else:
                return ""
        else:
            return ""

    def getStringUnchanged(self):
        if self.sourceNode and self.getValue:
            return str(self.getValue())
        else:
            return "[undefined]"

    def getStringRounded(self):
        if self.sourceNode and self.getValue:
            value = self.getValue()
            if value:
                return str(round(value, 3))
            else:
                return ""
        else:
            return "[undefined]"

    def getValueNodeName(self):
        if self.cache is not None:
            return self.cache
        else:
            s = ""
            item = self.sourceNode.item
            if item is None:
                return s
            for t in self.parentFields:
                # find tree
                tree = self.sourceNode
                while tree.parent is not None:
                    tree = tree.parent
                tree = tree.children[t]
                # find node in tree
                path = item.trees[t]
                node = tree.findNode(path)
                if node is not None:
                    parent = node.parent
                    if parent is not None:
                        s += parent.name
            return s

    def getValueNodePath(self):
        if self.cache is not None:
            return self.cache
        else:
            s = ""
            item = self.sourceNode.item
            if item is None:
                return s
            for t in self.parentFields:
                # find tree
                tree = self.sourceNode
                while tree.parent is not None:
                    tree = tree.parent
                tree = tree.children[t]
                # find node in tree
                path = item.trees[t]
                node = tree.findNode(path)
                if node and node.parent:
                    parent = node.parent
                    s += parent.name
                    while parent.parent and parent.parent.parent and parent.parent.parent.parent : # don't display forest or tree names
                        parent = parent.parent
                        s = parent.name + " | " + s
            return s

    def getValueString(self):
        values = self.getFieldValues()
        s = ""
        for v in values:
            if v:
                s += str(v)
        return s

    def getValueSum(self):
        values = self.getFieldValues()
        sum = 0
        for v in values:
            if v:     # either sum up using the value
                sum += v
            else:     # or the neutral element for addtion (0)
                sum += 0
        return sum

    def getValueDifference(self):
        """
        Calculates different a - b - c - d of values a,b,c,d
        """
        values = self.getFieldValues()
        difference = 0
        first = True
        for v in values:
            if first:
                if v:
                    difference = v     # first element is positive
                else:
                    difference = 0     # or the neutral element for addtion (0)
                first = False
            else:
                if v:
                    difference -= v     # all other elements are negative
                else:
                    difference += 0     # or the neutral element for addtion (0)
        return difference

    def getValueMean(self):
        values = self.getFieldValues()
        sum = 0.0
        n = 0.0
        for v in values:
            if v:     # in case of 'None', use neutal elements of 0 for addition, and 1 for multiplication
                n += 1.0
                sum += v
        if n > 0.0:
            return sum/n
        else:
            return None

    def getValueRatio(self):
        """
        Returns the ration a/(b+c+d+e+...) of field values a,b,c,d,e,...
        """
        values = self.getFieldValues()
        if len(values) < 2:
            return None
        else:
            denom = values[0]
            sum = 0
            for v in values[1:]:
                if v:
                    sum += v
            if sum != 0:
                return denom/sum
            else:
                return None
    
    def writeToString(self):
        string = "    field-type " + json.dumps(self.fieldType) + "\n"
        string += "        own-fields " + json.dumps(self.ownFields) + "\n"
        string += "        child-fields " + json.dumps(self.childFields) + "\n"
        string += "        sibling-fields " + json.dumps(self.siblingFields) + "\n"
        string += "        parent-fields " + json.dumps(self.parentFields) + "\n"
        return string

    def readFromString(self, string):
        
        s = string.split("\n        field-type ")
        name = json.loads(s[0])
        s = s[1].split("\n        own-fields ")
        try:
            self.fieldType = json.loads(s[0])
        except:
            self.printReadError(name, "field-type", s[0]);
        s = s[1].split("\n        child-fields ")
        try:
            self.ownFields = json.loads(s[0])
        except:
            self.printReadError(name, "own-fields", s[0]);
        s = s[1].split("\n        sibling-fields ")
        try:
            self.childFields = json.loads(s[0])
        except:
            self.printReadError(name, "child-fields", s[0]);
        s = s[1].split("\n        parent-fields ")
        try:
            self.siblingFields = json.loads(s[0])
        except:
            self.printReadError(name, "sibling-fields", s[0]);
        try:
            self.parentFields = json.loads(s[1])
        except:
            self.printReadError(name, "parent-fields", s[0]);

        self.initFieldType()
        
        return name

    def printReadError(self, name, parameter, string):
        print('error reading field parameter "' + parameter + '" of field "' + name + '": definition string "' + string + '" cannot be parsed.')


class Node:
    """
    A tree structure consisting of nodes that are parents of nodes etc. A node is a view-object to display one item.
    """
    
    def __init__(self, parent, tree, path):
        self.parent = parent
        self.children = []
        self.item = None
        self.name = ""
        self.fields = {}
        self.fieldOrder = []
        self.tree = tree
        self.path = path
        self.viewNode = None
        self.nameChangeCallback = None
        self.fieldChangeCallback = None
        self.deletionCallback = None
        self.moveCallback = None

    @staticmethod
    def _wrap_lines(raw_lines):
        """
        Helper function to wrap long text
        :param string: Input string, in one single line
        :return: Array with text wrapped at 70 chars, preserving newlines, but trimming space (no double new lines)
        """
        lines = []
        for line in [s.strip() for s in raw_lines.split('\n') if s.strip()]:
            lines += wrap(line, 70)
        return lines

    def map(self, function, parameter, depthFirst):
        ''' Applies the function to each node in the tree. The function must receive one parameter and return one parameter. The return value is used as parameter for the next function call. The value Parameter is used in the first call.'''
        
        # if depthFirst then recurse first and apply later
        if depthFirst:
            for c in self.children:
                parameter = c.map(function, parameter, depthFirst)
                
            return function(self, parameter)
        else:
            parameter = function(self, parameter)
            for c in self.children:
                parameter = c.map(function, parameter, depthFirst)
                
            return parameter

    def printForest(self, indent=0):
        
        # create leading white space
        s = ""
        for i in range(indent):
            s += "    "
        s += " → " + str(self.tree)
        for p in self.path:
            s += "[" + str(p) + "]"
        s += "    "
        if self.item is not None:
            for p in self.item.trees[self.tree]:
                s += "[" + str(p) + "]"
        s += " " + self.__class__.__name__ + " "
        
        # append item content and print
        s += self.name
        for name,field in self.fields.items():
            s += " [" + name + "] " + str(field.getString())
        print(s)
        
        # recurse
        for i in range(len(self.children)):
            self.children[i].printForest(indent + 1)

    def to_txt(self, lastitem=[]):
        """
        Create a text representation of the current branch.
        :param lastitem: A list of booleans showing whether the great-grandparent, grandparent, parent, is the
                         last item of the list of siblings
        :return: a string like this:

                    █ abc
                      def

                    █ abc
                    │ def
                    │
                    ┝━━█ abc
                    │  │ def
                    │  │
                    │  ┝━━█ abc
                    │  │    def
                    │  │
                    │  ┕━━█ abc
                    │       def
                    │
                    ┕━━█ abc
                       │ def
                       │
                       ┝━━█ abc
                       │    def
                       │
                       ┕━━█ abc
                            def
        """

        text = ""

        # create leading tree graphics
        # please pay attention: The spaces in the graphic strings are special unicode figure spaces
        if self.children:
            pre_node_prefix = ""
            first_line_prefix = "█  "
            line_prefix = "│  "
        else:
            pre_node_prefix = ""
            first_line_prefix = "█  "
            line_prefix = "   "
        n = 0
        for last in lastitem[::-1]:
            if not last:
                if n:
                    pre_node_prefix = "│  " + pre_node_prefix
                    first_line_prefix = "│  " + first_line_prefix
                    line_prefix = "│  " + line_prefix
                else:
                    pre_node_prefix = "│  " + pre_node_prefix
                    first_line_prefix = "├──" + first_line_prefix
                    line_prefix = "│  " + line_prefix
            else:
                if n:
                    pre_node_prefix = "   " + pre_node_prefix
                    first_line_prefix = "   " + first_line_prefix
                    line_prefix = "   " + line_prefix
                else:
                    pre_node_prefix = "│  " + pre_node_prefix
                    first_line_prefix = "└──" + first_line_prefix
                    line_prefix = "   " + line_prefix
            n += 1

        # append item content and print
        text += pre_node_prefix + "\n"

        # add node name (possibly multi-line)
        first = True
        for line in Node._wrap_lines(self.name):
            if first:
                text += first_line_prefix + line + "\n"
                first = False
            else:
                text += line_prefix + "    " + line + "\n"

        for name, field in self.fields.items():

            # wrap field content, larger bits of text start with a newline
            lines = Node._wrap_lines(field.getString())
            if len(lines) > 1:
                lines = [''] + lines

            # assemble final text with tree decorations
            first = True
            for line in lines:
                if first:
                    text += line_prefix + name + ": " + line + "\n"
                    first = False
                else:
                    text += line_prefix + "    " + line + "\n"

        # recurse
        sorted_children = sorted(self.children, key=lambda c: c.name)
        if len(sorted_children) > 1:
            childprefix = lastitem.copy()
            childprefix.append(False)
            for i in range(len(sorted_children)-1):
                text += sorted_children[i].to_txt(childprefix)
        if len(sorted_children):
            childprefix = lastitem.copy()
            childprefix.append(True)
            text += sorted_children[-1].to_txt(childprefix)

        # return
        return text

    def createPathTo(self, item, treeindex, nodeindex, viewtemplate):
        """
        Links an item into a tree, using the given path. Empty nodes get created on the way.
        """
        
        # either recurse deeper
        if nodeindex < len(item.trees[treeindex]):
            n = item.trees[treeindex][nodeindex]
            while n >= len(self.children):
                self.addChild()
            self.children[n].createPathTo(item,treeindex,nodeindex+1,viewtemplate)
        
        # or, if in final node, link item
        else:
            self.item = item
            self.name = item.name
            self.initFields(viewtemplate)
            self.registerCallbacks()

    def registerCallbacks(self):
        """
        Register callbacks from the parent (QNode), so changes from the tree layer
        can be shown in the QT GUI layer
        """
        if self.item is not None:
            self.item.registerNameChangeCallback(self.tree, self.notifyNameChange)
            self.item.registerFieldChangeCallback(self.tree, lambda x: self.notifyFieldChange(x, True))
            self.item.registerDeletionCallback(self.tree, self.notifyDeletion)
            self.item.registerMoveCallback(self.tree, self.notifyMove)
            self.item.registerSelectionCallback(self.tree, lambda x: self.notifySelection(x))
            self.item.registerViewNode(self.tree, self)

    def registerViewNode(self, viewNode):
        """
        Register the respective QT GUI view node with the node
        """
        self.viewNode = viewNode

    def findNode(self, path, currentIndex):
        """
        Recursive function to walk down a path in a tree and return the node at the last step
        """
        
        # either recurse deeper
        if currentIndex < len(path):
            childNumber = path[currentIndex]
            if childNumber >= len(self.children):
                return None
            return self.children[childNumber].findNode(path, currentIndex+1)
        
        # or, return myself
        else:
            return self

    def addChild(self):
        """
        Add a child to a node. The child is a new copy of the default node.
        """
        node = Node(self, self.tree, self.path + [len(self.children)])
        self.children += [node]
        return node

    def addNodeAsChild(self, node):
        """
        Add an existing node to a new parent.
        """
        self.children += [node]
        node.parent = self
        self.renumberChildren()
        node.item.notifyFieldChange("")
        self.notifyNameChange(self.name)

    def removeChild(self, child):
        if child in self.children:
            self.children.remove(child)
            self.renumberChildren()
            for f in self.fields:
                self.notifyFieldChange(f, True)

    def renumberChildren(self):
        """
        Correct the paths after children have been removed or added.
        """
        for i, c in enumerate(self.children):
            c.path = self.path + [i]
            if c.item:
                # print("{} {}".format(c.path, c.item.name))
                c.item.trees[self.tree] = self.path + [i]
                c.renumberChildren()

    def addItemAsChild(self, item):
        """
        Creates a node and links it to the item. Updates the item's forest indexes.
        """
        node = self.addChild()
        node.item = item
        node.name = item.name
        node.registerCallbacks()
        item.trees[node.tree] = node.path
        node.initFields(self.fields)
        return node

    def initFields(self, fields):
        self.fields = copy.deepcopy(fields)
        
        # first set all fields to myself
        for name, field in self.fields.items():
            field.sourceNode = self
        
        # and only then send notification
        for name, field in self.fields.items():
            self.notifyFieldChange(name, True)

    def findNodeByName(self, name):
        """
        Finds the first node of a given name in the tree.
        """
        if self.name == name:
            return self
        else:
            for c in self.children:
                found = c.findNodeByName(name)
                if found is not None:
                    return found
        return None

    def registerNameChangeCallback(self, callback):
        self.nameChangeCallback = callback

    def registerFieldChangeCallback(self, callback):
        self.fieldChangeCallback = callback

    def registerDeletionCallback(self, callback):
        self.deletionCallback = callback

    def registerMoveCallback(self, callback):
        self.moveCallback = callback

    def registerSelectionCallback(self, callback):
        self.selectionCallback = callback

    def notifyNameChange(self, newName):
        """
        Callback used to notify a node of a name change. Changes the name, then
        recreates strings for the fiels 'node-name' and 'node-path'.
        Recurses down the tree to update the node-path strings of all children by calling the function
        notifyParentNameChange.
        """

        # if name is not false, set new name and notify GUI
        if newName:
            self.name = newName
            if self.nameChangeCallback is not None:
                self.nameChangeCallback(newName)

        # always adapt all my fields in all trees (better: call the callback from the item without recursion)
        if self.item:
            for v in self.item.viewNodes:
                v and v.notifyFieldChange(False, False)

        # notify all children, recursive, but only update their fields
        for c in self.children:
            c.notifyNameChange(False)

    def notifyFieldChange(self, fieldName, recursion):
        """
        Callback, called whenever a field in a related item has changed.
        """

        if self.fieldChangeCallback is not None:
            for f in self.fields:
                self.fieldChangeCallback(f, self.fields[f].getString())
        
        if recursion:
            if self.parent is not None:
                # notify all siblings, non-recursive
                for c in self.parent.children:
                    if c is not self:
                        c.notifyFieldChange(fieldName, False)
                # notify all parents, recursive
                self.parent.notifyFieldChange(fieldName, True)

    def notifyDeletion(self):
        """
        Callback, called when the underlying item was deleted or removed from this tree.
        Recursion removes the complete child branch in this tree.
        """

        # unlink item
        self.item = None
        
        # tell the GUI layer to remove my QNode from its parent
        if self.deletionCallback is not None:
            self.deletionCallback()
            
        # recurse down my children
        for c in self.children:
            c.item.removeFromTree(self.tree)

        # make parent renumber children
        self.parent.removeChild(self)

    def notifyMove(self):
        """
        Callback, called when the underlying item moved to a different parent within the tree.
        Recursion renumbers all children in the new and in the old parent.
        """

        # find new parent and remember old parent
        oldParent = self.parent
        tree = self
        while tree.parent.parent is not None:
            tree = tree.parent
        path = self.item.trees[self.tree]
        newParent = tree.findNode(path[0:-1])

        # move node
        oldParent.removeChild(self)
        newParent.addNodeAsChild(self)

        # tell the GUI layer to move my QNode to its new parent
        if self.moveCallback is not None:
            self.moveCallback()

    def notifySelection(self, select):
        """
        Callback, called whenever the underlying item was selected.
        """

        # tell the GUI layer to select my QNode
        if self.selectionCallback is not None:
            self.selectionCallback(select)

    def removeEmptyNodes(self):
        """
        Recurse to find nodes that have no items. If found, those nodes are removed and their siblings renumbered.
        Called after initial loading of the file.
        """
        for c in self.children:
            c.removeEmptyNodes()
        if not self.item:
            self.notifyDeletion()


class Tree(Node):
    """
    A tree inside a forest. One item can appear several times in the forest, but only once in each tree.
    """

    def __init__(self, parent, index):
        """Initialise"""
        
        super().__init__(parent, index, [])
        self.fields = {}
        self.name = ""

    def createPathTo(self, item, treeindex):
        """
        Sort the item into the forest, creating existing nodes on the fly if missing.
        """

        # per tree: loop over all nodes, creating if necessary
        # per node: create final node and link it to item
        if item.trees[treeindex]:
            n = item.trees[treeindex][0]
            while n >= len(self.children):
                self.addChild()
            self.children[n].createPathTo(item, treeindex, 1, self.fields)

    def findNode(self, path):
        """
        Sort the item into the forest, creating existing nodes on the fly if missing.
        """

        # per tree: loop over all nodes, creating if necessary
        return super().findNode(path, 0)

    def writeToString(self):
        string = "tree " + json.dumps(self.name) + "\n"
        for n,f in self.fields.items():
            string += "    field " + json.dumps(n) + "\n"
            string += "    " + f.writeToString()
        return string

    def readFromString(self, string):
        fieldStrings = string.split("\n    field ")
        for n,fs in enumerate(fieldStrings):
            if n==0:
                self.name = json.loads(fs)
            else:
                f = Field()
                name = f.readFromString(fs)
                self.fields[name] = f
                self.fieldOrder += [name]

    def removeEmptyNodes(self):
        """
        Recurse to find nodes that have no items. If found, those nodes are removed and their siblings renumbered.
        Overrides same function in node, does not remove self (trees have no item anyway).
        """
        for c in self.children:
            c.removeEmptyNodes()


class Forest(Node):
    """
    The trunk node containing trees, that contain the nodes. Also manages the node templates.
    """

    def __init__(self, filename):
        """Initialise"""
        
        super().__init__(None, None, [])
        self.itemPool = None
        self.itemTypes = None
        self.readFromFile(filename)

    def createPaths(self):
        """
        Sort all items from the itempool into the forest,
        creating the forest structure as defined by the items' node indexes
        """
        for item in self.itemPool.items:
            self.createPathTo(item)

    def createPathTo(self, item):
        
        # start: loop over all trees, creating if necessary
        for b in range(len(item.trees)):
            if b >= len(self.children):
                self.addTree()
            if item.trees[b]:
                self.children[b].createPathTo(item,b)

    def addTree(self):
        self.children += [Tree(self, len(self.children))]

    def renumberChildren(self):
        """
        Correct the paths after children have been removed or added.
        """
        for i, c in enumerate(self.children):
            c.renumberChildren()

    def removeEmptyNodes(self):
        """
        Recurse to find nodes that have no items. If found, those nodes are removed and their siblings renumbered.
        Overrides same function in node, does not remove self (forests have no item anyway).
        """
        for c in self.children:
            c.removeEmptyNodes()

    def writeToString(self):
        s = ""
        for b in self.children:
            s += b.writeToString() + "\n"
        return s

    def readFromString(self, string):
        string = string.split("\n\ntree ")
        n = 0
        for s in string:
            if s != "":
                self.children[n].readFromString(s)
                n += 1

    def writeToFile(self, filename):
        treeString = self.writeToString()
        typeString = self.itemTypes.writeToString()
        itemString = self.itemPool.writeToString()
        with open(filename, "w") as f:
            f.write("--trees--\n\n")
            f.write(treeString)
            f.write("--item-types--\n\n")
            f.write(typeString)
            f.write("--item-pool--\n\n")
            f.write(itemString)

    def readFromFile(self, filename):
        with open(filename, "r") as f:

            # chop up string into trees, types, and items part
            s = f.read()
            s = s.split("--trees--")[1]
            s = s.split("--item-types--")
            treeString = s[0]
            s = s[1].split("--item-pool--")
            typeString = s[0]
            itemString = s[1]

        # create item pool
        self.itemTypes = ItemPool()
        self.itemTypes.readFromString(typeString)

        # create type pool
        self.itemPool = ItemPool()
        self.itemPool.readFromString(itemString)
        self.createPaths()

        # create trees
        self.readFromString(treeString)
        self.createPaths()     # atm we still need this twice. Fix it.

        # remove empty nodes
        self.removeEmptyNodes()



#
# Tis file is part of TreeTime, a tree editor and data analyser
#
# Copyright (C) GPLv3, 2015, Jacob Kanev
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

# -*- coding:utf-8 -*-

import tree
import item

# create four single items and test read/write
a = item.Item("a", '{"description": {"content": "Item A", "type": "string"} }','[[0],[],[0]]')

b = item.Item("b", '{"description": {"content": "Item B", "type": "string"} }','[[0,0],[],[1]]')

c = item.Item("c", '{}', '[[0,1],[],[]]')
c.addField("description", {"type":"string", "content":"Item C"})
c.addField("amount", {"type":"integer", "content":4})

d = item.Item("d", "{}","[[0,0,0],[],[]]")
d.addField("description", {"type":"string", "content":"Item D"})
d.addField("amount", {"type":"integer", "content":1})

e = item.Item("e", "{}","[[0,0,1],[],[]]")
e.addField("description", {"type":"string", "content":"Item E"})
e.addField("amount", {"type":"integer", "content":2})

x = item.Item("x", '{"description": {"content": "default node", "type": "string"}, "amount": {"content": 23, "type": "integer"} }','[[],[],[]]')

print("creating tree with amounts in d,e,c:")
print("      a      ")
print("     / \     ")
print("    b   c(4) ")
print("   /  \      ")
print(" d(1)  e(2)  ")

p = item.ItemPool()
p.items = [x,a,b,c,d,e]
p.defaultItem = x

# central tree object with no parent and no children
print("creating tree:")
tree_p = tree.Forest("items.data")
tree_p.printForest()

# add new items as child of node e
tree_p.children[2].addItemAsChild( p.addNewItem() )
tree_p.children[2].addItemAsChild( p.addNewItem() )
tree_p.children[2].addItemAsChild( p.addNewItem() )
tree_p.children[2].addItemAsChild( p.addNewItem() )

print("writing pool to file and reading it back in")
tree_p.writeToFile("items2.data")

tree_q = tree.Forest("items2.data")
# central tree object with no parent and no children
print("creating tree:")
tree_q.printForest()





#
# Tis file is part of TreeTime, a tree editor and data analyser
#
# Copyright (C) GPLv3, 2015, Jacob Kanev
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

# -*- coding:utf-8 -*-

#!/usr/bin/python3

import sys
from .item import *
from .tree import *
from .mainwindow import *
import datetime
import os.path
import platform
from PyQt5 import QtCore, QtGui, QtWidgets, uic
from threading import Timer
from PyQt5.QtGui import QPalette, QColor, QIcon
from pkg_resources import resource_filename

# Use only for debugging purposes (to cause an error on purpose, if you feel there might be loops), can cause segfaults
# sys.setrecursionlimit(50)


class QNode(QtWidgets.QTreeWidgetItem):
    """
    The GUI counterpart of a node. Displays the contents of a node.
    """
    
    def __init__(self, sourceNode, fieldOrder):
        
        # initialise display
        self.sourceNode = sourceNode
        displayStrings = [sourceNode.name]
        for d in fieldOrder:
            if d in self.sourceNode.fields:
                displayStrings += [self.sourceNode.fields[d].getString()]
            else:
                displayStrings += [""]
                
        super().__init__(displayStrings)
        
        # build reverse field order dictionary
        self.fieldOrder = {}
        for i,f in enumerate(fieldOrder):
            self.fieldOrder[f] = i+1     # plus one because column 0 is the node name
        
        # recurse
        for c in sourceNode.children:
            child = QNode(c, fieldOrder)
            super().addChild(child)
        
        # register callbacks
        self.registerCallbacks()

    def parent(self):
        """
        Returns the parent of a node
        """
        return super().parent() or self.treeWidget().invisibleRootItem() or None

    def registerCallbacks(self):
        """
        Registers QNode-specific callbacks with the source node
        """
        self.sourceNode.registerNameChangeCallback(self.notifyNameChange)
        self.sourceNode.registerFieldChangeCallback(self.notifyFieldChange)
        self.sourceNode.registerDeletionCallback(self.notifyDeletion)
        self.sourceNode.registerSelectionCallback(lambda x: self.notifySelection(x))
        self.sourceNode.registerMoveCallback(self.notifyMove)
        self.sourceNode.registerViewNode(self)

    def notifyNameChange(self, newName):
        super().setText(0, newName)

    def notifyFieldChange(self, fieldName, fieldContent):
        if fieldName in self.fieldOrder:
            super().setText(self.fieldOrder[fieldName], fieldContent)

    def notifyDeletion(self):
        # unlink node
        self.sourceNode = None
        
        # unlink from parent
        if self.parent() is not None:
            self.parent().removeChild(self)
        else:
            print("I have no parent, so I'll stay.")

    def notifyMove(self):

        # unlink from parent
        self.parent().removeChild(self)
        newParent = self.sourceNode.parent.viewNode
        newParent.addChild(self)

    def notifySelection(self, select):
        self.setSelected(select)


class UrlWidget(QtWidgets.QWidget):
    """
    Special custom widget class for URL fields
    """

    def __init__(self, url, callback, parent=None):
        """
        Initialise
        """

        # init
        QtWidgets.QWidget.__init__(self, parent=parent)
        self.url = url
        self.callback = callback

        # create line edit control and open button
        linewidget = QtWidgets.QLineEdit(url)
        linewidget.textChanged.connect(self.textChanged)
        openbutton = QtWidgets.QPushButton("Open")
        openbutton.clicked.connect(self.buttonClicked)

        # put them next to each other
        layout = QtWidgets.QHBoxLayout(self)
        layout.addWidget(openbutton)
        layout.addWidget(linewidget)

    def textChanged(self, text):
        """
        Callback, to save the new text and notify the parent that the text has changed
        """
        self.url = text
        self.callback()

    def toPlainText(self):
        """
        Callback, called by the GUI to retrieve (possibly changed) URL
        """
        return self.url

    def buttonClicked(self):
        """
        Callback, called when the button has been clicked. Opens the URL with the system
        default software.
        """
        QtGui.QDesktopServices.openUrl(QtCore.QUrl(self.url))


class TextEdit(QtWidgets.QPlainTextEdit):
    """
    Special custom widget class for URL fields
    """


    def __init__(self, text, callback, parent=None):
        """
        Initialise
        """

        # init
        QtWidgets.QPlainTextEdit.__init__(self, text, parent=parent)
        self.callback = callback
        self.has_changed = False
        self.textChanged.connect(self.notifyChange)

    def notifyChange(self):
        self.has_changed = True

    def focusOutEvent(self, e: QtGui.QFocusEvent):
        """
        Callback, to save the new text and notify the parent that the text has changed
        """
        if self.has_changed:
            self.has_changed = False
            self.callback()
        super().focusOutEvent(e)

    def leaveEvent(self, e: QtCore.QEvent):
        """
        Callback, to save the new text and notify the parent that the text has changed
        """
        if self.has_changed:
            self.has_changed = False
            self.callback()
        super().leaveEvent(e)


class TimerWidget(QtWidgets.QWidget):
    """
    A widget displaying an hour timer - a time display (hours:minutes:seconds)
    and a button "start"/"stop"
    """

    timer = None

    def __init__(self, partial, running_since, callback, parent=None):
        """
        initialise the timer object
        """

        # init
        QtWidgets.QWidget.__init__(self, parent=parent)
        self.start = running_since and datetime.datetime.strptime(running_since, "%Y-%m-%d %H:%M:%S")  # time when the timer was started last, or None if not running
        self.partial = partial or 0.0     # stored time span + time since timer was started
        self.total = self.partial
        if self.start:
            buttontext = "Stop"
            linetext = "(running)"
        else:
            buttontext = "Start"
            linetext = str(self.partial)
        self.callback = callback

        # create line edit control and open button
        self.linewidget = QtWidgets.QLineEdit(linetext)
        self.linewidget.textEdited.connect(self.textEdited)
        self.startstopbutton = QtWidgets.QPushButton(buttontext)
        self.startstopbutton.clicked.connect(self.buttonClicked)

        # put them next to each other
        layout = QtWidgets.QHBoxLayout(self)
        layout.addWidget(self.startstopbutton)
        layout.addWidget(self.linewidget)

        # start timer if it is loaded as running from file
        if self.start:
            self.updateValue()

    def textEdited(self, text):
        """
        Callback, to save the new text and notify the parent that the text has changed
        """
        if self.start:
            self.linewidget.setText("stop watch running")
        else:
            if text:
                try:
                    d = json.loads(text)
                except json.JSONDecodeError:
                    d = 0.0
            else:
                d = 0.0
            self.total = d
            self.partial = self.total
            self.callback()

    def toPlainText(self):
        """
        Callback, called by the GUI to retrieve (possibly changed) total value
        """
        return str(self.total)

    def runningSince(self):
        """
        Callback, returns the time the timer was started as YYYY-MM-DD hh.mm.ss string
        """
        return self.start and self.start.strftime("%Y-%m-%d %H:%M:%S") or False

    def buttonClicked(self):
        """
        Callback, called when the button has been clicked. Opens the URL with the system
        default software.
        """
        if self.start:
            self.stopTimer()
        else:
            self.startTimer()

    def startTimer(self):
        """
        Starts the timer
        """

        # save the time the timer was started
        self.start = datetime.datetime.now()
        self.startstopbutton.setText("Stop")
        self.linewidget.setText("stop watch running")

        # trigger the first update
        self.updateValue()

    def stopTimer(self):
        """
        Stops the timer
        """
        if self.start:

            # stop the timer, remove possible running timer objects
            self.startstopbutton.setText("Start")

            # store total value
            self.updateValue(update=False)
            self.partial = round(self.total*10000.0) / 10000.0
            self.total = self.partial
            self.start = None
            self.linewidget.setText(str(self.total))

            # notify back-end
            self.callback()

    def updateValue(self, update=True):
        """
        Updates the value in the GUI once and triggers the next update
        Used for sequencial updates while the timer is running. Updates are triggered using a threading.Timer object
        and occur every five seconds
        """

        # first, update the total time
        elapsed = datetime.datetime.now() - self.start
        self.total = self.partial \
                     + elapsed.days * 24.0 \
                     + elapsed.seconds / 60.0 / 60.0 \
                     + elapsed.microseconds / 60.0 / 1000000.0
        if update:
            self.callback()


class TreeTimeWindow(QtWidgets.QMainWindow, Ui_MainWindow):
    """
    Implements the main part of the GUI.
    """

    def __init__(self, filename=None):
        """
        Initialise the application, connect all button signals to application functions, load theme and last file
        """

        # initialise main window
        super().__init__()
        self.setupUi(self)

        # connect all button signals
        self.pushButtonNewChild.clicked.connect(lambda: self.createNode("child", False))
        self.pushButtonNewSibling.clicked.connect(lambda: self.createNode("sibling", False))
        self.pushButtonNewParent.clicked.connect(lambda: self.createNode("parent", False))
        self.pushButtonCopyNodeChild.clicked.connect(lambda: self.createNode("child", True))
        self.pushButtonCopyNodeSibling.clicked.connect(lambda: self.createNode("sibling", True))
        self.pushButtonCopyNodeParent.clicked.connect(lambda: self.createNode("parent", True))
        self.pushButtonCopyBranchSibling.clicked.connect(lambda: self.createNode("sibling", True, True))
        self.pushButtonNewFromTemplate.clicked.connect(self.pushButtonNewFromTemplateClicked)
        self.pushButtonLoadFile.clicked.connect(self.pushButtonLoadFileClicked)
        self.pushButtonSaveToFile.clicked.connect(self.pushButtonSaveToFileClicked)
        self.pushButtonExportBranchTxt.clicked.connect(self.pushButtonExportBranchTxtClicked)
        self.pushButtonExportTreeTxt.clicked.connect(self.pushButtonExportTreeTxtClicked)
        self.pushButtonRemove.clicked.connect(lambda: self.moveCurrentItemToNewParent(self.currentTree, None))
        self.pushButtonDelete.clicked.connect(self.pushButtonDeleteClicked)
        self.tableWidget.cellChanged.connect(self.tableWidgetCellChanged)
        self.tableWidget.verticalHeader().setSectionResizeMode(3)
        self.tableWidget.horizontalHeader().setSectionResizeMode(0, 2)     # column 0: fixed
        self.tableWidget.horizontalHeader().setSectionResizeMode(1, 2)     # column 1: fixed, 100
        self.tableWidget.horizontalHeader().resizeSection(1, 100)
        self.tableWidget.horizontalHeader().setSectionResizeMode(2, 2)     # column 2: fixed
        self.tableWidget.horizontalHeader().setSectionResizeMode(3, 1)     # column 3: stretch
        self.tableWidget.horizontalHeader().setSectionResizeMode(4, 2)     # column 4: fixed
        self.tabWidget.currentChanged.connect(self.tabWidgetCurrentChanged)
        self.write_delay = 0
        self.write_timer = False
        self.locked = True
        self.auto_updates = []
        self.update_timer = QtCore.QTimer(self)
        self.update_timer.timeout.connect(self.updateTimers)
        self.update_timer.start(1000)

        # init application settings
        self.settings = QtCore.QSettings('FreeSoftware', 'TreeTime')

        # load last file
        lastFile = self.settings.value('lastFile')
        if lastFile:
            self.loadFile(lastFile)
            self.setWindowTitle("TreeTime - " + lastFile)
            self.settings.setValue('fileDir', os.path.dirname(lastFile))
            self.settings.setValue('lastFile', lastFile)
            self.labelCurrentFile.setText(lastFile)
        else:
            self.pushButtonLoadFileClicked()

        # init themes and set last theme
        self.system_palette = self.palette()
        self.fillThemeBox()
        self.fillColourBox()
        self.cboxThemeTextChanged()
        self.cboxColoursTextChanged()
        self.cboxTheme.currentTextChanged.connect(self.cboxThemeTextChanged)
        self.cboxColours.currentTextChanged.connect(self.cboxColoursTextChanged)

        # show window
        self.showMaximized()

    def closeEvent(self, event):
        """
        Do some cleanup before you close the window
        """

        # Write state to file
        if self.write_timer:
            self.write_timer.cancel()
        self.write_delay = 0
        self.writeToFile()

        # cancel all running stop watches
        # \todo

        # continue closing the window
        event.accept()  # let the window close

    def fillThemeBox(self):
        """
        Fills the theme selection box with all themes the system is capable of
        """
        for k in QtWidgets.QStyleFactory.keys():
            self.cboxTheme.addItem(k)
        current = self.settings.value('theme')
        if current:
            self.cboxTheme.setCurrentText(current)

    def fillColourBox(self):
        """
        Fills the theme selection box with all themes the system is capable of
        """
        self.cboxColours.addItem("Dark")
        self.cboxColours.addItem("Light")
        self.cboxColours.addItem("System")
        current = self.settings.value('colours')
        if current:
            self.cboxColours.setCurrentText(current)

    def cboxThemeTextChanged(self):
        """
        Callback from the theme selector combo box. Sets a new theme and stores it in the settings.
        """
        application = QtWidgets.QApplication.instance()
        style = self.cboxTheme.currentText()
        self.settings.setValue('theme', style)
        application.setStyle(QtWidgets.QStyleFactory.create(style))
        self.treeSelectionChanged(self.currentTree)

    def cboxColoursTextChanged(self):
        """
        Callback from the theme selector combo box. Sets a new theme and stores it in the settings.
        """
        application = QtWidgets.QApplication.instance()
        colour = self.cboxColours.currentText()
        self.settings.setValue('colours', colour)
        palette = QPalette()

        # Apply dark colours
        if colour == "Dark":
            palette.setColor(QPalette.Window, QColor(50, 52, 54))
            palette.setColor(QPalette.WindowText, QtCore.Qt.white)
            palette.setColor(QPalette.Base, QColor(40, 42, 44))
            palette.setColor(QPalette.AlternateBase, QColor(50, 52, 54))
            palette.setColor(QPalette.ToolTipBase, QtCore.Qt.black)
            palette.setColor(QPalette.ToolTipText, QtCore.Qt.white)
            palette.setColor(QPalette.Text, QtCore.Qt.white)
            palette.setColor(QPalette.Disabled, QPalette.Text, QColor(150, 150, 150))
            palette.setColor(QPalette.Button, QColor(60, 62, 64))
            palette.setColor(QPalette.ButtonText, QtCore.Qt.white)
            palette.setColor(QPalette.BrightText, QtCore.Qt.red)
            palette.setColor(QPalette.Link, QColor(60, 200, 255))
            palette.setColor(QPalette.Highlight, QColor(60, 200, 255))
            palette.setColor(QPalette.HighlightedText, QtCore.Qt.black)

        # apply light colours
        elif colour == "Light":
            palette = QPalette()
            palette.setColor(QPalette.Window, QColor(255-53, 255-53, 255-53))
            palette.setColor(QPalette.WindowText, QtCore.Qt.black)
            palette.setColor(QPalette.Base, QColor(255-25, 255-25, 255-25))
            palette.setColor(QPalette.AlternateBase, QColor(255-53, 255-53, 255-53))
            palette.setColor(QPalette.ToolTipBase, QtCore.Qt.white)
            palette.setColor(QPalette.ToolTipText, QtCore.Qt.black)
            palette.setColor(QPalette.Text, QtCore.Qt.black)
            palette.setColor(QPalette.Disabled, QPalette.Text, QColor(70, 70, 70))
            palette.setColor(QPalette.Button, QColor(255-53, 255-53, 255-53))
            palette.setColor(QPalette.ButtonText, QtCore.Qt.black)
            palette.setColor(QPalette.BrightText, QtCore.Qt.blue)
            palette.setColor(QPalette.Link, QColor(255-42, 255-130, 255-218))
            palette.setColor(QPalette.Highlight, QColor(255-42, 255-130, 255-218))
            palette.setColor(QPalette.HighlightedText, QtCore.Qt.white)

        # apply system colours
        elif colour == "System":
            palette = self.system_palette

        application.setPalette(palette)
        self.treeSelectionChanged(self.currentTree)

    def pushButtonSaveToFileClicked(self):
        """
        Callback for the save-file button. Saves the current data to a new file and keeps that file connected.
        """
        fileDir = self.settings.value('fileDir') or ''
        file = QtWidgets.QFileDialog.getSaveFileName(self, "Save new Data File", fileDir, '*.trt')[0]
        if file != '':
            self.labelCurrentFile.setText(file)
            self.writeToFile()
            self.setWindowTitle("TreeTime - " + file)
            self.settings.setValue('fileDir', os.path.dirname(file))
            self.settings.setValue('lastFile', file)

    def pushButtonLoadFileClicked(self):
        """
        Callback for the load-file button. Loads new file and keeps that file connected.
        """
        fileDir = self.settings.value('fileDir') or ''
        file = QtWidgets.QFileDialog.getOpenFileName(self, "Load Data File", fileDir, 'TreeTime Files (*.trt)')[0]
        if file != '':
            self.loadFile(file)
            self.setWindowTitle("TreeTime - " + file)
            self.settings.setValue('fileDir', os.path.dirname(file))
            self.settings.setValue('lastFile', file)
            self.labelCurrentFile.setText(file)

    def pushButtonExportBranchTxtClicked(self):
        """
        Callback for the txt export. Asks for a file name, then writes branch text export into it.
        """
        if self.currentItem:
            fileDir = self.settings.value('fileDir') or ''
            file = QtWidgets.QFileDialog.getSaveFileName(self, "Export to Plain Text", fileDir, '*.txt')[0]
            if file != '':
                with open(file, "w") as f:
                    currentNode = self.currentItem.viewNodes[self.currentTree]
                    txt = currentNode.to_txt()
                    f.write(txt)

    def pushButtonExportTreeTxtClicked(self):
        """
        Callback for the txt export. Asks for a file name, then writes branch text export into it.
        """
        if self.currentItem:
            fileDir = self.settings.value('fileDir') or ''
            file = QtWidgets.QFileDialog.getSaveFileName(self, "Export to Plain Text", fileDir, '*.txt')[0]
            if file != '':
                with open(file, "w") as f:
                    rootNode = self.forest.children[self.currentTree]
                    children = sorted(rootNode.children, key=lambda a: a.name)
                    for c in children:
                        f.write('\n')
                        f.write(c.to_txt())
                        f.write('\n')

    def pushButtonNewFromTemplateClicked(self):
        """
        Callback for the new-from-template button. Loads new file and immediately saves it to a different file.
        """

        # First load template file
        templateDir = self.settings.value('templateDir') or ''
        template = QtWidgets.QFileDialog.getOpenFileName(self, "Load Template", templateDir, '*.trt')[0]
        if template != '':
            self.loadFile(template)
            self.settings.setValue('templateDir', os.path.dirname(template))

            # Then save as new file
            self.pushButtonSaveToFileClicked()

    def pushButtonDeleteClicked(self):
        message = "Deleting node \"" + self.currentItem.name + "\". \n\n" \
                  "This will remove all descendents (children, grandchildren, ...) from the tree \"" \
                  + self.forest.children[self.currentTree].name +"\".\n" \
                  "Changes are saved to file immediately and cannot be reverted."
        msgBox = QtWidgets.QMessageBox(QtWidgets.QMessageBox.Warning, "Tree Time Message", message)
        msgBox.setStandardButtons(QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
        result = msgBox.exec_()
        if result == QtWidgets.QMessageBox.Ok:

            # remove running timers
            for f in self.currentItem.fields.keys():
                if self.currentItem.fields[f]['type'] == 'timer':
                    self.currentItem.fields[f]['running_since'] = False
                    self.adjustAutoUpdate(self.currentItem, f)

            # delete item and update file
            self.forest.itemPool.deleteItem(self.currentItem)
            self.delayedWriteToFile()
        
    def loadFile(self, filename):
        self.removeBranchTabs()
        self.tabWidgets = []
        self.treeWidgets = []
        self.currentTree = 0
        self.currentItem = None
        self.gridInitialised = False
        if filename is not None and filename != '':
            try:
                # load file
                self.forest = Forest(filename)
                self.createBranchTabs()
                self.fillTreeWidgets()

                # select first item
                if len(self.treeWidgets):
                    firstTree = self.treeWidgets[0]
                    items = firstTree.topLevelItemCount()
                    if items:
                        firstTree.topLevelItem(0).setSelected(True)

            # in case of failure, show user message and reload
            except KeyError as e:
                msgBox = QtWidgets.QMessageBox(QtWidgets.QMessageBox.Warning, "Tree Time Message", str(e))
                msgBox.setStandardButtons(QtWidgets.QMessageBox.Ok)
                result = msgBox.exec_()
                self.pushButtonLoadFileClicked()
        else:
            self.pushButtonLoadFileClicked()

    def delayedWriteToFile(self, countdown=False):
        """
        Schedules a write after 5 seconds. A timer is called that stops after 1 second and decreases a counter.
        If the counter is at zero, a write-to-file is performed. If the counter is > 0, it is counted down and a
        new timer is started.
        When changing cell contents, each cell change calls a delayed write. The file is only written if there has
        been no change for 5 seconds (otherwise the continuous file writing slows the user input down).
        :param countdown: Whether to count down and trigger a write at zero (=True),
                          or whether to init a new run (=False)
        :return:
        """

        # We've been called by a timer process, count down and possibly write
        if countdown:
            if self.write_delay > 0:
                self.write_delay -= 1
                if self.write_delay == 0:
                    self.writeToFile()
                self.write_timer = False

        # We've been called after a cell change, init the counter
        else:
            self.write_delay = 5

        # The counter is up, start another 2-second timer
        if self.write_delay > 0 and not self.write_timer:
            self.write_timer = Timer(1, self.delayedWriteToFile, kwargs={'countdown': True})
            self.write_timer.start()

    def writeToFile(self, delayed=False, countdown=False):
        self.forest.writeToFile(self.labelCurrentFile.text())

    def removeBranchTabs(self):
            self.tabWidget.clear()
            self.tabWidgets = []

    def createBranchTabs(self):
        
        self.locked = True
        
        # create tabs and tree widgets
        for n,c in enumerate(self.forest.children):
            newTab = QtWidgets.QWidget()
            newTab.setObjectName("tab"+str(n))
            newGridLayout = QtWidgets.QGridLayout(newTab)
            self.tabWidget.addTab(newTab, "")
            self.tabWidgets += [newTab]
            self.tabWidget.setTabText(n, c.name)
            newTree = QtWidgets.QTreeWidget(newTab)
            newTree.setGeometry(QtCore.QRect(0, 0, 731, 751))
            newTree.setAllColumnsShowFocus(True)
            newTree.setWordWrap(True)
            newTree.headerItem().setText(0, "1")
            newTree.setColumnCount(len(c.fieldOrder))
            newGridLayout.addWidget(newTree, 0, 0, 1, 1)
            self.treeWidgets += [newTree]
        
        self.locked = False

    def fillTreeWidgets(self):
        """
        Fills all tree tabs by creating all nodes
        """

        # for each tree in the forest
        for n, c in enumerate(self.forest.children):
            self.treeWidgets[n].itemSelectionChanged.connect(lambda x=n: self.treeSelectionChanged(x))
            self.treeWidgets[n].setHeaderLabels([""] + c.fieldOrder)
            root = self.treeWidgets[n].invisibleRootItem()
            c.viewNode = root

            # add branch
            for b in c.children:
                parent = QNode(b, c.fieldOrder)
                root.addChild(parent)
            # expand name column so all names are readable
            self.treeWidgets[n].resizeColumnToContents(0)

            # init sorting
            self.treeWidgets[n].setSortingEnabled(True)
            self.treeWidgets[n].sortItems(0, QtCore.Qt.AscendingOrder)

    def _protectCells(self, row, columns):

        if not self.gridInitialised:
            nonEditFlags = QtCore.Qt.ItemFlags()
            empties = []
            for i in columns:
                if not self.tableWidget.item(row, i):
                    empties.append(QtWidgets.QTableWidgetItem(""))
                    empties[-1].setFlags(nonEditFlags)
                    self.tableWidget.setItem(row, i, empties[-1])
                else:
                    self.tableWidget.item(row, i).setFlags(nonEditFlags)

    def treeSelectionChanged(self, treeIndex):
        
        if not self.locked:
            
            self.locked = True

            # init
            selectedItems = self.treeWidgets[treeIndex].selectedItems()

            # we have something to write
            if selectedItems != []:

                qnode = selectedItems[0]
                
                # unselect previous item in all trees
                if self.currentItem is not None:
                    try:
                        self.currentItem.select(False)
                    except:
                        print("warning: Unselecting of {} didn't work".format(self.currentItem.name))
                        pass

                # select current item in all trees
                self.currentItem = qnode.sourceNode.item
                if self.currentItem is None:
                    self.tableWidget.clear()
                    self.gridInitialised = False
                    self.locked = False
                    return
                self.currentItem.select(True)
                
                # expand current item in current tree
                parent = qnode.parent()
                while parent is not None:
                    parent.setExpanded(True)
                    parent = parent.parent()
                
                # create non-edit flags
                nonEditFlags = QtCore.Qt.ItemFlags()

                # go through all lines of the table
                n = 0

                # empty line
                self._protectCells(0, [0, 1, 2, 3, 4])
                n += 1

                # add name
                name = QtWidgets.QTableWidgetItem("")
                name.setFlags(nonEditFlags)
                value = QtWidgets.QTableWidgetItem(self.currentItem.name)
                font = name.font()
                font.setPointSize(font.pointSize() + 3)
                value.setFont(font)
                self.tableWidget.setItem(n, 1, name)
                self.tableWidget.setItem(n, 3, value)
                self._protectCells(n, [0, 2, 4])
                n += 1
                
                # empty line
                self._protectCells(n, [0, 1, 2, 3, 4])
                n += 1

                # add all parents in tree parent path
                for treeNumber, path in enumerate(self.currentItem.trees):
                    tree = self.forest.children[treeNumber]
                    name = QtWidgets.QTableWidgetItem(tree.name)
                    name.setFlags(nonEditFlags)
                    name.setTextAlignment(0x82)
                    self.tableWidget.setItem(n, 1, name)
                    buttonbox = QtWidgets.QDialogButtonBox()
                    if not len(path):
                        button = QtWidgets.QToolButton()
                        button.setText("")
                        button.setToolButtonStyle(1)  # 1 - text only
                        button.setFixedWidth(button.sizeHint().height())
                        button.setFixedHeight(button.sizeHint().height())
                        button.setPopupMode(QtWidgets.QToolButton.InstantPopup)     # no down arrow is shown, menu pops up on click
                        button.setStyleSheet("::menu-indicator{ image: none; }")
                        button.setMenu(self.createParentMenu(treeNumber, self.forest))
                        buttonbox.addButton(button, QtWidgets.QDialogButtonBox.ButtonRole.ResetRole)
                    elif len(path) == 1:
                        parent = tree.findNode(path).parent
                        button = QtWidgets.QToolButton()
                        button.setArrowType(4)     # 4 - rightarrow
                        button.setToolButtonStyle(1)     # get square buttons of same size as text buttons for > buttons
                        button.setStyleSheet("::menu-indicator{ image: none; }")
                        button.setFixedWidth(button.sizeHint().height())
                        button.setFixedHeight(button.sizeHint().height())
                        button.setToolButtonStyle(0)     # 0 - icon only
                        button.setPopupMode(QtWidgets.QToolButton.InstantPopup)     # no down arrow is shown, menu pops up on click
                        button.setMenu(self.createParentMenu(treeNumber, parent))
                        buttonbox.addButton(button, QtWidgets.QDialogButtonBox.ButtonRole.ResetRole)
                    else:
                        for p in range(1, len(path)):
                            parent = tree.findNode(path[0:p])
                            button = QtWidgets.QToolButton()
                            button.setArrowType(4)     # 4 - rightarrow
                            button.setText(parent.name)
                            button.setToolButtonStyle(1)     # 2 - text beside icon
                            button.setPopupMode(QtWidgets.QToolButton.InstantPopup)     # no down arrow is shown, menu pops up on click
                            button.setStyleSheet("::menu-indicator{image:none;}")
                            button.setMenu(self.createParentMenu(treeNumber, parent.parent))
                            buttonbox.addButton(button, QtWidgets.QDialogButtonBox.ButtonRole.ResetRole)
                        button = QtWidgets.QToolButton()
                        button.setArrowType(4)     # 4 - rightarrow
                        button.setToolButtonStyle(1)     # get square buttons of same size as text buttons for > buttons
                        button.setStyleSheet("::menu-indicator{ image: none; }")
                        button.setFixedWidth(button.sizeHint().height())
                        button.setFixedHeight(button.sizeHint().height())
                        button.setToolButtonStyle(0)     # 0 - icon only
                        button.setPopupMode(QtWidgets.QToolButton.InstantPopup)     # no down arrow is shown, menu pops up on click
                        button.setMenu(self.createParentMenu(treeNumber, parent))
                        buttonbox.addButton(button, QtWidgets.QDialogButtonBox.ButtonRole.ResetRole)

                    self.tableWidget.setCellWidget(n, 3, buttonbox)
                    self._protectCells(n, [0, 2, 4])
                    n += 1
                    
                # empty line
                self._protectCells(n, [0, 1, 2, 3, 4])
                n += 1

                # add item fields
                for key in sorted(self.currentItem.fields):
                    name = QtWidgets.QTableWidgetItem(key)
                    name.setFlags(nonEditFlags)
                    name.setTextAlignment(0x82)
                    if self.currentItem.fields[key]['type'] == 'text':
                        text = self.currentItem.fields[key]["content"]
                        text = text and str(text) or ""     # display "None" values as empty string
                        widget = TextEdit(text, lambda row=n: self.tableWidgetCellChanged(row, 3))
                        self.tableWidget.setCellWidget(n, 3, widget)
                    elif self.currentItem.fields[key]['type'] == 'url':
                        value = self.currentItem.fields[key]["content"]
                        value = value and str(value) or ""  # display "None" values as empty string
                        widget = UrlWidget(value, lambda row=n: self.tableWidgetCellChanged(row, 3))
                        self.tableWidget.setCellWidget(n, 3, widget)
                    elif self.currentItem.fields[key]['type'] == 'timer':
                        value = self.currentItem.fields[key]["content"]
                        running_since = self.currentItem.fields[key]["running_since"]
                        widget = TimerWidget(value, running_since, lambda row=n: self.tableWidgetCellChanged(row, 3))
                        self.tableWidget.setCellWidget(n, 3, widget)
                    else:
                        value = self.currentItem.fields[key]["content"]
                        value = value and str(value) or ""     # display "None" values as empty string
                        value = QtWidgets.QTableWidgetItem(value)
                        self.tableWidget.setItem(n, 3, value)
                    self.tableWidget.setItem(n, 1, name)
                    self._protectCells(n, [0, 2, 4])
                    n += 1

            # an empty page, clear an initialise
            else:
                self.tableWidget.clear()
                self.gridInitialised = False
                n = 0

            # empty lines to fill the 20 lines in the main view
            if n < 20:
                for k in range(n, 20):
                    self._protectCells(k, [0, 1, 2, 3, 4])

            self.gridInitialised = True
            self.locked = False
    
    def createParentMenu(self, treeIndex, parent):
        """
        Displays a menu with possible children to select, at the current mouse cursor position.
        """
        menu = QtWidgets.QMenu()

        # root node
        if not parent.parent:
            action = QtWidgets.QAction("Add to Top Level", menu)
            action.triggered.connect(lambda checked, t=treeIndex,
                                            p=self.forest.children[treeIndex]: self.moveCurrentItemToNewParent(t, p))
            menu.addAction(action)
            parent = self.forest.children[treeIndex]
            menu.addSeparator()

        # top level node only
        else:
            if not parent.parent.parent:
                action = QtWidgets.QAction("Remove from Tree", menu)
                action.triggered.connect(lambda checked, t=treeIndex, p=False: self.moveCurrentItemToNewParent(t, p))
                menu.addAction(action)
                action = QtWidgets.QAction("Add to Top Level", menu)
                action.triggered.connect(lambda checked, t=treeIndex, p=self.forest.children[treeIndex]:
                                         self.moveCurrentItemToNewParent(t, p))
                menu.addAction(action)
                parent = self.forest.children[treeIndex]
                menu.addSeparator()

        # all other nodes
        currentNode = self.currentItem.viewNodes[treeIndex]
        for c in sorted(parent.children, key=lambda x: x.name):
            if c != currentNode:
                action = QtWidgets.QAction(c.name, menu)
                action.triggered.connect(lambda checked, t=treeIndex, p=c: self.moveCurrentItemToNewParent(t, p))
                menu.addAction(action)

        return menu

    def tabWidgetCurrentChanged(self, tree):
        '''Called when the user selects another tree in the tab widget.
        '''

        self.currentTree = tree
        self.treeSelectionChanged(tree)

    def tableWidgetCellChanged(self, row, column):
        """
        Called when the user wants to change the item name, field content or parent via the grid
        """

        if not self.locked:
            self.locked = True
            if column == 3:
                
                # the node name has been changed
                if row == 1:
                    newName = self.tableWidget.item(row, column).text()
                    self.currentItem.changeName(newName)
                
                # one of the fields has been changed
                else:
                    fieldName = self.tableWidget.item(row, 1).text()
                    fieldType = self.currentItem.fields[fieldName]['type']
                    if fieldType in ('text', 'url'):
                        newValue = self.tableWidget.cellWidget(row, 3).toPlainText()
                    elif fieldType == 'timer':
                        newValue = (self.tableWidget.cellWidget(row, 3).toPlainText(),
                                    self.tableWidget.cellWidget(row, 3).runningSince())
                    else:
                        newValue = self.tableWidget.item(row, 3).text()
                    result = self.currentItem.changeFieldContent(fieldName, newValue)
                    if result is not True:
                        message = "Couldn't update field content.\n" + str(result)
                        msgBox = QtWidgets.QMessageBox(QtWidgets.QMessageBox.Warning, "Tree Time Message", message)
                        msgBox.exec_()

                    # add/remove timer to list
                    if fieldType == 'timer':
                        self.adjustAutoUpdate(self.currentItem, fieldName)

            self.locked = False
            self.delayedWriteToFile()

    def adjustAutoUpdate(self, item, fieldName):
        """ Adds or removes a timer to the list of fields to auto-update.
        """
        if item.fields[fieldName]['type'] == 'timer':
            # if timer is running, add it to the list
            if item.fields[fieldName]['running_since']:
                if (item, fieldName) not in self.auto_updates:
                    self.auto_updates.append((item, fieldName))

            # if timer is stopped, remove it from the list
            else:
                if (item, fieldName) in self.auto_updates:
                    del self.auto_updates[self.auto_updates.index((item, fieldName))]

    def updateTimers(self):
        # update all timers that are running
        for item, fieldName in self.auto_updates:
            result = item.changeFieldContent(fieldName, [True])
        # remove timers that are problematic

    def createNode(self, insertas, copy, recurse = False, srcItem = None, destItem = None):
        
        treeWidget = self.treeWidgets[self.currentTree]

        # set source and destination nodes and qnodes
        if recurse and srcItem is not None and destItem is not None:
            sourceItem = srcItem
            sourceNode = srcItem.viewNodes[self.currentTree]
            sourceQNode = sourceNode.viewNode
            destNode = destItem.viewNodes[self.currentTree]
            destQNode = destNode.viewNode
        elif len(treeWidget.selectedItems()):
            sourceQNode = treeWidget.selectedItems()[0]
            sourceNode = sourceQNode.sourceNode
            sourceItem = sourceNode.item
            destNode = sourceNode.parent
            destQNode = sourceQNode.parent()
        else:
            sourceQNode = treeWidget.invisibleRootItem()
            sourceNode = self.forest.children[self.currentTree]
            sourceItem = sourceNode.item
            destQNode = sourceQNode
            destNode = sourceNode
            insertas = 'root'

        if destQNode is None:
            destQNode = treeWidget.invisibleRootItem()

        if copy:
            item = self.forest.itemPool.copyItem(sourceItem)
            for n,t in enumerate(self.forest.children):
                if n != self.currentTree and item.trees[n] != []:
                    oldNode = t.findNode(item.trees[n])
                    newNode = oldNode.parent.addItemAsChild(item)
                    newQNode = QNode(newNode, self.forest.children[n].fieldOrder)
                    parent = oldNode.viewNode.parent()
                    if parent is None:
                        parent = self.treeWidgets[n].invisibleRootItem()
                    parent.addChild(newQNode)
        else:
            # add new item by copying the first type
            item = self.forest.itemPool.copyItem(self.forest.itemTypes.items[0])

        if insertas == "root":

            # create default node and add item to it
            node = sourceNode.addItemAsChild(item)
            qnode = QNode(node, self.forest.children[self.currentTree].fieldOrder)
            sourceQNode.addChild(qnode)

            # expand new entry
            qnode.setExpanded(True)

        if insertas == "child":

            # create default node and add item to it
            node = sourceNode.addItemAsChild(item)
            qnode = QNode(node, self.forest.children[self.currentTree].fieldOrder)
            sourceQNode.addChild(qnode)

            # expand parent and select new item
            sourceQNode.setExpanded(True)

        elif insertas == "sibling":

            if destNode is None or destQNode is None:
                return

            # create default node and add item to it
            node = destNode.addItemAsChild(item)
            qnode = QNode(node, self.forest.children[self.currentTree].fieldOrder)
            destQNode.addChild(qnode)

            # expand parent and select new item
            destQNode.setExpanded(True)

        elif insertas == "parent":

            if destNode is None or destQNode is None:
                return

            node = destNode.addItemAsChild(item)

            # move original node to be child of new parent
            destNode.removeChild(sourceNode)
            node.addNodeAsChild(sourceNode)
            destQNode.removeChild(sourceQNode)
            qnode = QNode(node, self.forest.children[self.currentTree].fieldOrder)
            destQNode.addChild(qnode)

            # expand parent and select new item
            qnode.setExpanded(True)

        if recurse:
            for c in sourceNode.children:
                self.createNode(insertas, copy, recurse, c.item, node.item)

        if srcItem is None:
            treeWidget.setCurrentItem(qnode)
            self.delayedWriteToFile()
    
    def moveCurrentItemToNewParent(self, treeIndex, newParent):
        treeWidget = self.treeWidgets[self.currentTree]
        if len(treeWidget.selectedItems()):

            self.locked = True

            # save current item
            item = self.currentItem

            # error message if recursion is tried
            if newParent and (newParent.item == item):
                message = "You are trying to set\n" + item.name + "\nto be its own parent. This is not supported."
                msgBox = QtWidgets.QMessageBox(QtWidgets.QMessageBox.Warning, "Tree Time Message", message)
                msgBox.setStandardButtons(QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel);
                return msgBox.exec_()

            # find old parent
            oldParent = item.viewNodes[treeIndex]
            disappeared = False

            # either remove node from tree
            if oldParent and not newParent:

                # question to user: Really remove?
                message = "Removing node \"" + item.name + "\" from the tree \"" \
                          + self.forest.children[self.currentTree].name +"\".\n\n" \
                          "This will remove all descendents (children, grandchildren, ...). " \
                          "Changes are saved to file immediately and cannot be reverted."
                msgBox = QtWidgets.QMessageBox(QtWidgets.QMessageBox.Warning, "Tree Time Message", message)
                msgBox.setStandardButtons(QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
                result = msgBox.exec_()
                
                # remove if user has confirmed
                if result == QtWidgets.QMessageBox.Ok:
                    item.removeFromTree(treeIndex)
                    if treeIndex == self.currentTree:
                        disappeared = True     # remember that we've removed a node from the current tree
                else:
                    return

            # or move node within the tree
            elif oldParent and newParent and newParent.item:
                item.moveInTree(treeIndex, newParent.item.trees[treeIndex])
            elif oldParent and newParent and not newParent.item:
                item.moveInTree(treeIndex, [])

            # or add node to tree
            elif not oldParent and newParent:
                tree = self.forest.children[treeIndex]
                node = newParent.addItemAsChild(item)
                newQNode = QNode(node, tree.fieldOrder)
                newParent.viewNode.addChild(newQNode)
            
            # save change
            item.notifyFieldChange('')

            # select new item after moving, if it is still part of the current tree
            if not disappeared:
                treeWidget.setCurrentItem(item.viewNodes[self.currentTree].viewNode)

            # clear table widget to prevent segfault crash in qt lib
            self.locked = False
            self.gridInitialised = False
            self.tableWidget.clear()

            # update table and write to file
            self.treeSelectionChanged(self.currentTree)
            self.delayedWriteToFile()


class TreeTime:
    
    def __init__(self):
        
        app = QtWidgets.QApplication(sys.argv)
        if platform.system() == "Windows":
            app.setStyle("Fusion")

        # add application icon
        icon_file = os.path.dirname(os.path.realpath(__file__))
        icon_file += os.path.sep + '..' + os.path.sep + 'data' + os.path.sep + 'treetime-logo.png'
        icon_file = resource_filename('treetime', 'data/treetime-logo.png')
        print(icon_file)
        main_window = TreeTimeWindow()
        main_window.setWindowIcon(QIcon(icon_file))
        # tray_icon = QtWidgets.QSystemTrayIcon(QIcon(file), parent=app)
        main_window.show()
        # tray_icon.show()
        sys.exit(app.exec_())





import os.path
from PyQt5 import QtCore, QtGui, QtWidgets, uic

# compile ui file if necessary
if os.path.getmtime("mainwindow.py") < os.path.getmtime("mainwindow.ui"):
    with open("mainwindow.py", "w") as file:
        print("compiling ui")
        uic.compileUi("mainwindow.ui", file)




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'mainwindow.ui'
#
# Created by: PyQt5 UI code generator 5.15.1
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(1448, 843)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.horizontalLayout = QtWidgets.QHBoxLayout(self.centralwidget)
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.toolBox = QtWidgets.QToolBox(self.centralwidget)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.toolBox.sizePolicy().hasHeightForWidth())
        self.toolBox.setSizePolicy(sizePolicy)
        self.toolBox.setMinimumSize(QtCore.QSize(210, 0))
        self.toolBox.setMaximumSize(QtCore.QSize(16777215, 16777215))
        self.toolBox.setObjectName("toolBox")
        self.fileIo = QtWidgets.QWidget()
        self.fileIo.setGeometry(QtCore.QRect(0, -125, 200, 840))
        self.fileIo.setObjectName("fileIo")
        self.verticalLayout = QtWidgets.QVBoxLayout(self.fileIo)
        self.verticalLayout.setObjectName("verticalLayout")
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.label_8 = QtWidgets.QLabel(self.fileIo)
        self.label_8.setWordWrap(True)
        self.label_8.setObjectName("label_8")
        self.verticalLayout.addWidget(self.label_8)
        self.labelCurrentFile = QtWidgets.QLabel(self.fileIo)
        self.labelCurrentFile.setEnabled(True)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.labelCurrentFile.sizePolicy().hasHeightForWidth())
        self.labelCurrentFile.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setItalic(True)
        self.labelCurrentFile.setFont(font)
        self.labelCurrentFile.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.labelCurrentFile.setWordWrap(True)
        self.labelCurrentFile.setOpenExternalLinks(True)
        self.labelCurrentFile.setTextInteractionFlags(QtCore.Qt.LinksAccessibleByMouse)
        self.labelCurrentFile.setObjectName("labelCurrentFile")
        self.verticalLayout.addWidget(self.labelCurrentFile)
        spacerItem1 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem1)
        self.label_5 = QtWidgets.QLabel(self.fileIo)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_5.setFont(font)
        self.label_5.setScaledContents(False)
        self.label_5.setAlignment(QtCore.Qt.AlignCenter)
        self.label_5.setObjectName("label_5")
        self.verticalLayout.addWidget(self.label_5)
        self.pushButtonNewFromTemplate = QtWidgets.QPushButton(self.fileIo)
        self.pushButtonNewFromTemplate.setObjectName("pushButtonNewFromTemplate")
        self.verticalLayout.addWidget(self.pushButtonNewFromTemplate)
        self.pushButtonLoadFile = QtWidgets.QPushButton(self.fileIo)
        self.pushButtonLoadFile.setObjectName("pushButtonLoadFile")
        self.verticalLayout.addWidget(self.pushButtonLoadFile)
        self.pushButtonSaveToFile = QtWidgets.QPushButton(self.fileIo)
        self.pushButtonSaveToFile.setObjectName("pushButtonSaveToFile")
        self.verticalLayout.addWidget(self.pushButtonSaveToFile)
        spacerItem2 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem2)
        self.label_11 = QtWidgets.QLabel(self.fileIo)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_11.setFont(font)
        self.label_11.setAlignment(QtCore.Qt.AlignCenter)
        self.label_11.setObjectName("label_11")
        self.verticalLayout.addWidget(self.label_11)
        self.pushButtonExportBranchTxt = QtWidgets.QPushButton(self.fileIo)
        self.pushButtonExportBranchTxt.setObjectName("pushButtonExportBranchTxt")
        self.verticalLayout.addWidget(self.pushButtonExportBranchTxt)
        self.pushButtonExportTreeTxt = QtWidgets.QPushButton(self.fileIo)
        self.pushButtonExportTreeTxt.setObjectName("pushButtonExportTreeTxt")
        self.verticalLayout.addWidget(self.pushButtonExportTreeTxt)
        spacerItem3 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem3)
        self.label_9 = QtWidgets.QLabel(self.fileIo)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_9.setFont(font)
        self.label_9.setAlignment(QtCore.Qt.AlignCenter)
        self.label_9.setObjectName("label_9")
        self.verticalLayout.addWidget(self.label_9)
        self.cboxTheme = QtWidgets.QComboBox(self.fileIo)
        self.cboxTheme.setObjectName("cboxTheme")
        self.verticalLayout.addWidget(self.cboxTheme)
        self.cboxColours = QtWidgets.QComboBox(self.fileIo)
        self.cboxColours.setObjectName("cboxColours")
        self.verticalLayout.addWidget(self.cboxColours)
        spacerItem4 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem4)
        spacerItem5 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem5)
        self.label_10 = QtWidgets.QLabel(self.fileIo)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_10.setFont(font)
        self.label_10.setAlignment(QtCore.Qt.AlignCenter)
        self.label_10.setObjectName("label_10")
        self.verticalLayout.addWidget(self.label_10)
        self.label_7 = QtWidgets.QLabel(self.fileIo)
        self.label_7.setObjectName("label_7")
        self.verticalLayout.addWidget(self.label_7)
        self.toolBox.addItem(self.fileIo, "")
        self.pageTree = QtWidgets.QWidget()
        self.pageTree.setGeometry(QtCore.QRect(0, 0, 200, 716))
        self.pageTree.setMinimumSize(QtCore.QSize(200, 716))
        self.pageTree.setObjectName("pageTree")
        self.formLayout = QtWidgets.QFormLayout(self.pageTree)
        self.formLayout.setFieldGrowthPolicy(QtWidgets.QFormLayout.ExpandingFieldsGrow)
        self.formLayout.setObjectName("formLayout")
        self.label = QtWidgets.QLabel(self.pageTree)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label.setFont(font)
        self.label.setAlignment(QtCore.Qt.AlignCenter)
        self.label.setObjectName("label")
        self.formLayout.setWidget(2, QtWidgets.QFormLayout.SpanningRole, self.label)
        self.pushButtonNewChild = QtWidgets.QPushButton(self.pageTree)
        self.pushButtonNewChild.setObjectName("pushButtonNewChild")
        self.formLayout.setWidget(3, QtWidgets.QFormLayout.SpanningRole, self.pushButtonNewChild)
        self.pushButtonNewSibling = QtWidgets.QPushButton(self.pageTree)
        self.pushButtonNewSibling.setObjectName("pushButtonNewSibling")
        self.formLayout.setWidget(4, QtWidgets.QFormLayout.SpanningRole, self.pushButtonNewSibling)
        self.pushButtonNewParent = QtWidgets.QPushButton(self.pageTree)
        self.pushButtonNewParent.setObjectName("pushButtonNewParent")
        self.formLayout.setWidget(5, QtWidgets.QFormLayout.SpanningRole, self.pushButtonNewParent)
        self.label_2 = QtWidgets.QLabel(self.pageTree)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_2.setFont(font)
        self.label_2.setAlignment(QtCore.Qt.AlignCenter)
        self.label_2.setObjectName("label_2")
        self.formLayout.setWidget(7, QtWidgets.QFormLayout.SpanningRole, self.label_2)
        self.pushButtonCopyNodeChild = QtWidgets.QPushButton(self.pageTree)
        self.pushButtonCopyNodeChild.setObjectName("pushButtonCopyNodeChild")
        self.formLayout.setWidget(8, QtWidgets.QFormLayout.SpanningRole, self.pushButtonCopyNodeChild)
        self.pushButtonCopyNodeSibling = QtWidgets.QPushButton(self.pageTree)
        self.pushButtonCopyNodeSibling.setObjectName("pushButtonCopyNodeSibling")
        self.formLayout.setWidget(9, QtWidgets.QFormLayout.SpanningRole, self.pushButtonCopyNodeSibling)
        self.pushButtonCopyNodeParent = QtWidgets.QPushButton(self.pageTree)
        self.pushButtonCopyNodeParent.setObjectName("pushButtonCopyNodeParent")
        self.formLayout.setWidget(10, QtWidgets.QFormLayout.SpanningRole, self.pushButtonCopyNodeParent)
        self.label_3 = QtWidgets.QLabel(self.pageTree)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_3.setFont(font)
        self.label_3.setAlignment(QtCore.Qt.AlignCenter)
        self.label_3.setObjectName("label_3")
        self.formLayout.setWidget(12, QtWidgets.QFormLayout.SpanningRole, self.label_3)
        self.label_6 = QtWidgets.QLabel(self.pageTree)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.label_6.sizePolicy().hasHeightForWidth())
        self.label_6.setSizePolicy(sizePolicy)
        self.label_6.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.label_6.setObjectName("label_6")
        self.formLayout.setWidget(13, QtWidgets.QFormLayout.LabelRole, self.label_6)
        self.pushButtonCopyBranchSibling = QtWidgets.QPushButton(self.pageTree)
        self.pushButtonCopyBranchSibling.setObjectName("pushButtonCopyBranchSibling")
        self.formLayout.setWidget(15, QtWidgets.QFormLayout.SpanningRole, self.pushButtonCopyBranchSibling)
        self.pushButtonCopyChildrenSiblings = QtWidgets.QPushButton(self.pageTree)
        self.pushButtonCopyChildrenSiblings.setObjectName("pushButtonCopyChildrenSiblings")
        self.formLayout.setWidget(16, QtWidgets.QFormLayout.SpanningRole, self.pushButtonCopyChildrenSiblings)
        self.label_4 = QtWidgets.QLabel(self.pageTree)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_4.setFont(font)
        self.label_4.setAlignment(QtCore.Qt.AlignCenter)
        self.label_4.setObjectName("label_4")
        self.formLayout.setWidget(18, QtWidgets.QFormLayout.SpanningRole, self.label_4)
        self.pushButtonRemove = QtWidgets.QPushButton(self.pageTree)
        self.pushButtonRemove.setObjectName("pushButtonRemove")
        self.formLayout.setWidget(19, QtWidgets.QFormLayout.SpanningRole, self.pushButtonRemove)
        self.pushButtonDelete = QtWidgets.QPushButton(self.pageTree)
        self.pushButtonDelete.setObjectName("pushButtonDelete")
        self.formLayout.setWidget(20, QtWidgets.QFormLayout.SpanningRole, self.pushButtonDelete)
        spacerItem6 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.formLayout.setItem(1, QtWidgets.QFormLayout.SpanningRole, spacerItem6)
        spacerItem7 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.formLayout.setItem(6, QtWidgets.QFormLayout.SpanningRole, spacerItem7)
        spacerItem8 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.formLayout.setItem(11, QtWidgets.QFormLayout.SpanningRole, spacerItem8)
        spacerItem9 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.formLayout.setItem(17, QtWidgets.QFormLayout.SpanningRole, spacerItem9)
        self.comboBox = QtWidgets.QComboBox(self.pageTree)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.comboBox.sizePolicy().hasHeightForWidth())
        self.comboBox.setSizePolicy(sizePolicy)
        self.comboBox.setObjectName("comboBox")
        self.comboBox.addItem("")
        self.comboBox.addItem("")
        self.comboBox.addItem("")
        self.comboBox.addItem("")
        self.comboBox.addItem("")
        self.comboBox.addItem("")
        self.comboBox.addItem("")
        self.comboBox.addItem("")
        self.comboBox.addItem("")
        self.comboBox.addItem("")
        self.comboBox.addItem("")
        self.formLayout.setWidget(13, QtWidgets.QFormLayout.FieldRole, self.comboBox)
        spacerItem10 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.formLayout.setItem(22, QtWidgets.QFormLayout.SpanningRole, spacerItem10)
        spacerItem11 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.formLayout.setItem(21, QtWidgets.QFormLayout.SpanningRole, spacerItem11)
        self.toolBox.addItem(self.pageTree, "")
        self.pageGlobal = QtWidgets.QWidget()
        self.pageGlobal.setGeometry(QtCore.QRect(0, 0, 96, 26))
        self.pageGlobal.setObjectName("pageGlobal")
        self.gridLayout_3 = QtWidgets.QGridLayout(self.pageGlobal)
        self.gridLayout_3.setObjectName("gridLayout_3")
        spacerItem12 = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_3.addItem(spacerItem12, 0, 0, 1, 1)
        self.toolBox.addItem(self.pageGlobal, "")
        self.horizontalLayout.addWidget(self.toolBox)
        self.tableWidget = QtWidgets.QTableWidget(self.centralwidget)
        self.tableWidget.setMidLineWidth(0)
        self.tableWidget.setShowGrid(False)
        self.tableWidget.setWordWrap(True)
        self.tableWidget.setRowCount(20)
        self.tableWidget.setColumnCount(5)
        self.tableWidget.setObjectName("tableWidget")
        self.tableWidget.horizontalHeader().setVisible(False)
        self.tableWidget.horizontalHeader().setCascadingSectionResizes(False)
        self.tableWidget.horizontalHeader().setDefaultSectionSize(10)
        self.tableWidget.horizontalHeader().setMinimumSectionSize(10)
        self.tableWidget.horizontalHeader().setSortIndicatorShown(True)
        self.tableWidget.horizontalHeader().setStretchLastSection(False)
        self.tableWidget.verticalHeader().setVisible(False)
        self.tableWidget.verticalHeader().setCascadingSectionResizes(True)
        self.horizontalLayout.addWidget(self.tableWidget)
        self.tabWidget = QtWidgets.QTabWidget(self.centralwidget)
        self.tabWidget.setDocumentMode(False)
        self.tabWidget.setObjectName("tabWidget")
        self.horizontalLayout.addWidget(self.tabWidget)
        self.horizontalLayout.setStretch(0, 1)
        self.horizontalLayout.setStretch(1, 3)
        self.horizontalLayout.setStretch(2, 5)
        MainWindow.setCentralWidget(self.centralwidget)
        self.actionQuit = QtWidgets.QAction(MainWindow)
        self.actionQuit.setObjectName("actionQuit")

        self.retranslateUi(MainWindow)
        self.toolBox.setCurrentIndex(0)
        self.tabWidget.setCurrentIndex(-1)
        self.actionQuit.triggered.connect(MainWindow.close)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "Tree Time"))
        self.label_8.setText(_translate("MainWindow", "Connected to Data File:"))
        self.labelCurrentFile.setText(_translate("MainWindow", "[current file name]"))
        self.label_5.setText(_translate("MainWindow", "Load and Save"))
        self.pushButtonNewFromTemplate.setText(_translate("MainWindow", "New from Template"))
        self.pushButtonLoadFile.setText(_translate("MainWindow", "Load File"))
        self.pushButtonSaveToFile.setText(_translate("MainWindow", "Save As"))
        self.label_11.setText(_translate("MainWindow", "Export"))
        self.pushButtonExportBranchTxt.setText(_translate("MainWindow", "Export Branch to Textfile"))
        self.pushButtonExportTreeTxt.setText(_translate("MainWindow", "Export Tree to Textfile"))
        self.label_9.setText(_translate("MainWindow", "Look and Feel"))
        self.label_10.setText(_translate("MainWindow", "About"))
        self.label_7.setText(_translate("MainWindow", "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0//EN\" \"http://www.w3.org/TR/REC-html40/strict.dtd\">\n"
"<html><head><meta name=\"qrichtext\" content=\"1\" /><style type=\"text/css\">\n"
"p, li { white-space: pre-wrap; }\n"
"</style></head><body style=\" font-family:\'Noto Sans\'; font-size:10pt; font-weight:400; font-style:normal;\">\n"
"<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">     TreeTime</p>\n"
"<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">     Version 2020.07</p>\n"
"<p style=\" margin-top:12px; margin-bottom:12px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px;\">     New versions,<br />     source code and <br />     documentation at <br />     <a href=\"https://github.com/jkanev/treetime\"><span style=\" text-decoration: underline; color:#2980b9;\">github.com/jkanev/treetime</span></a></p></body></html>"))
        self.toolBox.setItemText(self.toolBox.indexOf(self.fileIo), _translate("MainWindow", "File"))
        self.label.setText(_translate("MainWindow", "Insert New Nodes"))
        self.pushButtonNewChild.setText(_translate("MainWindow", "New Child"))
        self.pushButtonNewSibling.setText(_translate("MainWindow", "New Sibling"))
        self.pushButtonNewParent.setText(_translate("MainWindow", "New Parent"))
        self.label_2.setText(_translate("MainWindow", "Copy Existing Nodes"))
        self.pushButtonCopyNodeChild.setText(_translate("MainWindow", "Copy Node as Child"))
        self.pushButtonCopyNodeSibling.setText(_translate("MainWindow", "Copy Node as Sibling"))
        self.pushButtonCopyNodeParent.setText(_translate("MainWindow", "Copy Node as Parent"))
        self.label_3.setText(_translate("MainWindow", "Copy Parts of the Tree"))
        self.label_6.setText(_translate("MainWindow", "Depth  "))
        self.pushButtonCopyBranchSibling.setText(_translate("MainWindow", "Copy Branch as Sibling"))
        self.pushButtonCopyChildrenSiblings.setText(_translate("MainWindow", "Copy Children to Siblings"))
        self.label_4.setText(_translate("MainWindow", "Remove Single Nodes"))
        self.pushButtonRemove.setText(_translate("MainWindow", "Remove from this Tree"))
        self.pushButtonDelete.setText(_translate("MainWindow", "Delete Item"))
        self.comboBox.setItemText(0, _translate("MainWindow", "all levels"))
        self.comboBox.setItemText(1, _translate("MainWindow", "1 level"))
        self.comboBox.setItemText(2, _translate("MainWindow", "2 levels"))
        self.comboBox.setItemText(3, _translate("MainWindow", "3 levels"))
        self.comboBox.setItemText(4, _translate("MainWindow", "4 levels"))
        self.comboBox.setItemText(5, _translate("MainWindow", "5 levels"))
        self.comboBox.setItemText(6, _translate("MainWindow", "6 levels"))
        self.comboBox.setItemText(7, _translate("MainWindow", "7 levels"))
        self.comboBox.setItemText(8, _translate("MainWindow", "8 levels"))
        self.comboBox.setItemText(9, _translate("MainWindow", "9 levels"))
        self.comboBox.setItemText(10, _translate("MainWindow", "10 levels"))
        self.toolBox.setItemText(self.toolBox.indexOf(self.pageTree), _translate("MainWindow", "Edit (this Tree)"))
        self.toolBox.setItemText(self.toolBox.indexOf(self.pageGlobal), _translate("MainWindow", "Edit (all Trees)"))
        self.tableWidget.setSortingEnabled(False)
        self.actionQuit.setText(_translate("MainWindow", "Quit"))



#
# Tis file is part of TreeTime, a tree editor and data analyser
#
# Copyright (C) GPLv3, 2015, Jacob Kanev
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

# -*- coding:utf-8 -*-

import copy
import json
from threading import Timer

class Item:
    """
    The list/forest item containing the actual data
    """

    def __init__(self, name, fieldstring='{}', treestring='[]'):
        self.name = name
        self.parentNames = []
        self.fields = json.loads(fieldstring)
        self.trees = json.loads(treestring)
        self.viewNodes = []
        self.nameChangeCallbacks = []
        self.fieldChangeCallbacks = []
        self.deletionCallbacks = []
        self.moveCallbacks = []
        self.selectionCallbacks = []
        self.clearCallbacks()

    def __deepcopy__(self, memo):
        ''' Overload of the deepcopy function, to avoid copying the node recursively.'''
        
        if memo is None:
            memo = {}
        
        itemType = self.__class__
        newItem = itemType.__new__(itemType)
        
        # fields and trees are complex
        newItem.fields = copy.deepcopy(self.fields)
        newItem.trees = copy.deepcopy(self.trees)
        
        # names are simple
        newItem.name = self.name
        newItem.parentNames = self.parentNames
        
        # everything else should be blank
        newItem.viewNodes = []
        newItem.nameChangeCallbacks = []
        newItem.fieldChangeCallbacks = []
        newItem.deletionCallbacks = []
        newItem.moveCallbacks = []
        newItem.selectionCallbacks = []
        newItem.clearCallbacks()
        
        memo[id(newItem)] = newItem
        return newItem

    def __repr__(self):
        string = "name: {} | fields: {} | trees: {}".format(self.name,
                                                            json.dumps(self.fields),
                                                            json.dumps(self.trees))
        return string

    def clearCallbacks(self):
        self.viewNodes = []
        self.nameChangeCallbacks = []
        self.fieldChangeCallbacks = []
        self.deletionCallbacks = []
        self.moveCallbacks = []
        self.selectionCallbacks = []
        for t in self.trees:
            self.viewNodes += [None]
            self.nameChangeCallbacks += [None]
            self.fieldChangeCallbacks += [None]
            self.deletionCallbacks += [None]
            self.moveCallbacks += [None]
            self.selectionCallbacks += [None]

    def addField(self, name, content):
        self.fields[name] = content

    def removeField(self, name, content):
        del self.fields[name]

    def writeToString(self):
        string = self.name + "\n"
        string += "    fields " + json.dumps(self.fields) + "\n"
        string += "    trees " + json.dumps(self.trees) + "\n"
        return string

    def readFromString(self, string):
        s = string.split("\n    fields ")
        self.name = s[0]
        s = s[1].split("\n    trees ")
        self.fields = json.loads(s[0])
        self.trees = json.loads(s[1])
        self.clearCallbacks()

    def printitem(self):
        print(self.name)
        for key in self.fields:
            print("    ", key)
            for subkey in self.fields[key]:
                print("        ", subkey, ":", self.fields[key][subkey])

    def registerViewNode(self, tree, node):
        self.viewNodes[tree] = node

    def registerNameChangeCallback(self, tree, callback):
        self.nameChangeCallbacks[tree] = callback

    def registerSelectionCallback(self, tree, callback):
        self.selectionCallbacks[tree] = callback

    def registerFieldChangeCallback(self, tree, callback):
        self.fieldChangeCallbacks[tree] = callback

    def registerDeletionCallback(self, tree, callback):
        self.deletionCallbacks[tree] = callback

    def registerMoveCallback(self, tree, callback):
        self.moveCallbacks[tree] = callback

    def changeName(self, newName):
        self.name = newName
        for c in self.nameChangeCallbacks:
            if c is not None:
                c(newName)
    
    def select(self, select):
        for c in self.selectionCallbacks:
            if c is not None:
                c(select)

    def changeFieldContent(self, fieldName, fieldContent):
        """
        Edit the content of a field. The content is expected to be a string and
        will be converted according to the field type.
        """

        if fieldName not in self.fields:
            return "A field with name " + fieldName + " does not exist in node " + self.name + "."
        else:
            # update field content
            field = self.fields[fieldName]
            type = field["type"]
            if fieldContent:

                # changing the content of a string field - read the string plainly
                if type in ("string", "text", "url"):
                    field["content"] = fieldContent

                # changing the content of a number - read the string into a value
                elif type == "integer":
                    try:
                        field["content"] = json.loads(fieldContent)
                    except json.JSONDecodeError:
                        field["content"] = None

                # changing the content of a timer - real change comes in a tuple, mere value update just
                # with a "True" flag
                elif type == "timer":
                    try:
                        # standard content change, indicated by a tuple
                        if len(fieldContent) == 2:
                            field["content"] = json.loads(fieldContent[0])
                            field["running_since"] = fieldContent[1]

                        # just send an empty update message if a timer is running
                        else:
                            pass

                    except json.JSONDecodeError:
                        field["content"] = None
                        field["running_since"] = None
                else:
                    return "A field of type " + type + " cannot be edited."
            else:
                field["content"] = None

            # notify changes
            self.notifyFieldChange(fieldName)

        return True

    def notifyFieldChange(self, fieldName):
        """
        notify GUI of field change
        """
        for f in self.fieldChangeCallbacks:
            if f is not None:
                f(fieldName)

    def removeFromTree(self, treeIndex):
        """
        Remove this item from the tree, and have it call all removal callbacks.
        This triggers a recursion, deleting the whole child branch from the tree
        """

        self.trees[treeIndex] = []
        if self.deletionCallbacks[treeIndex] is not None:
            self.deletionCallbacks[treeIndex]()
            self.registerDeletionCallback(treeIndex, None)
            self.registerMoveCallback(treeIndex, None)
            self.registerFieldChangeCallback(treeIndex, None)
            self.registerNameChangeCallback(treeIndex, None)
            self.registerViewNode(treeIndex, None)

    def moveInTree(self, treeIndex, parentPath):
        """
        Move this item to a different parent at the new path
        """
        self.trees[treeIndex] = parentPath + [0]
        if self.moveCallbacks[treeIndex] is not None:
            self.moveCallbacks[treeIndex]()


class ItemPool:
    """
    A pool of items. Used to store all the tree nodes (items). The tree-related information is handled in the node
    classes.
    """

    def __init__(self):
        """
        Construct
        """
        self.items = []
        self.existing_paths = []     # all existing paths for each tree, used during loading to check for duplicates

    def writeToString(self):
        """
        Writes the whole pool to a re-loadable string
        """
        string = ""
        for it in sorted(self.items, key=lambda e: e.trees):
            string += "item " + it.writeToString() + "\n"
        return string
        
    def readFromString(self, string):
        """
        Reads the pool from a string
        """

        # split string and iterate over all parts
        string = string.split("\n\nitem ")     # items are separated by empty lines and the keyword "item"
        for s in string:
            if s != "":

                # read single item
                it = Item("")
                try:
                    it.readFromString(s)
                except json.decoder.JSONDecodeError as e:
                    raise KeyError("Corrupt data file. "
                                   "The item defined by {} in the file could not be read. "
                                   "Problem: {} when reading {}. "
                                   "Please correct manually in a text editor and then reload the file. "
                                   "".format(s, e.msg, e.doc))

                # sanity check: look for duplicated paths
                for i, t in enumerate(it.trees):
                    if len(self.existing_paths) <= i:
                        self.existing_paths.append([])
                    if t:

                        # if we found a duplicated path, raise an exception
                        if t in self.existing_paths[i]:
                            found_item = None
                            for f in self.items:
                                if f.trees[i] == t:
                                    found_item = f
                                    break
                            raise KeyError("Corrupt data file. "
                                           "The path {} in the tree {}, used by \"{}\" is already in use by \"{}\". "
                                           "Please correct manually in a text editor and then reload the file. "
                                           "".format(t, i, it.name, found_item.name))

                        # if all is fine, add item and proceed
                        else:
                            self.existing_paths[i].append(t)
                self.items += [it]

    def printpool(self):
        """
        Prints a list of items. This is a debug function.
        """
        for it in self.items:
            it.printitem()

    def copyItem(self, item):
        """
        Adds a copy of an item to the list and returns a reference to it
        """
        newitem = copy.deepcopy(item)
        newitem.viewNodes = []
        newitem.clearCallbacks()
        self.items += [newitem]
        return newitem

    def deleteItem(self, item):
        """
        Removes an item from the pool
        """
        for i, t in enumerate(item.trees):
            item.removeFromTree(i)
        self.items.remove(item)



#!/usr/bin/env python

from setuptools import setup

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='TreeTime',
    version='2021.2',
    description='TreeTime is a to-do list manager, test report tool, project manager, family ancestry editor,'
                'mind-mapping tool, etc. Using TreeTime you can categorise and organise any data in tree structures.'
                'You can define several trees, each with a different structure, on the same data. You can use functions'
                '(sums, ratios and means) recursively up the branches of a tree.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Jacob Kanev',
    author_email='jkanev@zoho.com',
    url='https://github.com/jkanev/treetime',
    packages=['treetime'],
    package_dir={'treetime': 'treetime'},
    include_package_data=True,
    entry_points={'gui_scripts': ['treetime = treetime.__main__:main']},
    package_data={'treetime': ['data/Simple-Task-List.trt', 'data/treetime-logo.png', 'docs/*.png']},
    # according to some people on the web install_requiring PyQt5 should work, but as my system gives me an error for
    # this, I'll comment it out. Please install it separately (pip3 PyQt5 apparently works)
    # install_requires=['PyQt5'],
)
